

""" 
Requires fusepy and fuse
- does not support win
"""
import os
import sys
import pathlib
import threading
import signal
import multiprocessing as mproc

from importlib import import_module
from fsspec import AbstractFileSystem
from fsspec.asyn import AsyncFileSystem
#from fsspec.fuse import FUSEr, FUSE

from lazy.libz import Lib
from typing import Dict, Type, Any, Optional, Union
from types import ModuleType
from lazy.models import BaseCls
from logz import get_logger

try: from fsspec.fuse import FUSEr, FUSE
except ImportError: FUSEr = FUSE = object

logger = get_logger('fuze')
FuseSystemType = Union[Type[AbstractFileSystem], Type[AsyncFileSystem]]

_lock: threading._RLock = threading._RLock()

class MountPoint(BaseCls):
    fuzer: str # Name of the class
    source: str # User provided Mount Source
    mount_point: str # User provided mount point
    cleanup: bool = True # Cleanup dir after unmount
    thread: Optional[threading.Thread] = None
    process: Optional[mproc.Process] = None
    alive: bool = True

    @property
    def source_path(self):
        uri_splits = self.source.split('://', maxsplit=1)
        if len(uri_splits) > 1: return uri_splits[-1].strip()
        return self.source

    @property
    def target_path(self):
        return pathlib.Path(self.mount_point).resolve()
    
    @property
    def target_path_str(self):
        return self.target_path.as_posix()

    def kill(self, timeout: int = 5):
        if self.thread: self.thread.join(timeout=timeout)
        elif self.process:
            os.kill(self.process.pid, signal.SIGTERM) 
            self.process.join(timeout=timeout)

        self.alive = False
        self.thread = None
        logger.info(f'[{self.fuzer}] Completed Unmount: {self.source}')
    
    def _background(self, fs: FuseSystemType, foreground: bool = True, threads: bool = False, ready_file: bool = False):
        self.alive = True
        from fsspec.fuse import FUSE as _FUSE
        from fsspec.fuse import FUSEr as _FUSEr
        func = lambda: _FUSE(
            _FUSEr(fs, self.source_path, ready_file=ready_file), 
            self.target_path_str, 
            nothreads=not threads, 
            foreground=foreground)
        
        while self.alive:
            try:
                logger.info(f'Started Mount {self.source} -> {self.target_path_str} in Background')
                func()
            except KeyboardInterrupt:
                logger.error('Got KeyboardInterrupt. Unmounting')
                self.kill()
                #_kill_proc(self.mount_point)
                break
            except Exception as e:
                logger.error(f'Got Exception {e}. Unmounting')
                self.kill()
                #_kill_proc(self.mount_point)
                break

    
    def mount(self, fs: FuseSystemType, foreground: bool = True, threads: bool = False, ready_file: bool = False, use_mp: bool = False, daemon: bool = True, ops_class: 'FUSEr' = None):
        #from fsspec.fuse import FUSE as _FUSE
        #from fsspec.fuse import FUSEr as _FUSEr
        #if not ops_class: ops_class = _FUSEr
        self.target_path.mkdir(parents=True, exist_ok=True)
        if use_mp:
            #with _lock.acquire():
            #with threading.Lock() as _lockz:
            _lock.acquire()
            self.process = mproc.Process(target=self._background, args=(fs, foreground, threads, ready_file,), daemon=daemon)
            self.process.start()
            _lock.release()

            #with threading.RLock() as _lock:
            #    _lock.acquire()
        elif not foreground:
            #with _lock.acquire():
            #with threading._RLock() as _lock:
            _lock.acquire()
            self.thread = threading.Thread(target=self._background, args=(fs, foreground, threads, ready_file,), daemon = daemon)
            self.thread.start()
            _lock.release()
        
        else:
            self._background(fs, foreground = foreground, threads = threads, ready_file = ready_file)
        



        """
        func = lambda: _FUSE(
            ops_class(fs, self.source_path, ready_file=ready_file), 
            self.target_path_str, 
            nothreads=not threads, 
            foreground=foreground)
        if not foreground:
            th = threading.Thread(target=func)
            th.daemon = True
            th.start()
            logger.info(f'Started Mount {self.source} -> {self.target_path_str} in Background')
            self.thread = th
            _add_proc(self.mount_point, self)

        else:  # pragma: no cover
            try:
                logger.info(f'Started Mount {self.source} -> {self.target_path_str} in Foreground')
                func()
            except KeyboardInterrupt:
                _kill_proc(self.mount_point)
                #sys.exit()
        """


    def unmount(self, timeout: int = 5):
        #if not self.alive and self.thread is None: return
        logger.info(f'[{self.fuzer}] Unmounting: {self.source} -> {self.target_path.as_posix()}')
        self.kill(timeout)
        if self.cleanup:
            logger.info(f'[{self.fuzer}] Cleaning Up Mount Point: {self.target_path.as_posix()}')
            self.target_path.unlink()



## Keep track of all mounted
_ALL_FUZES: Dict[str, MountPoint] = {}
_FUZE_READY: bool = False
_FUZE_ALLOWED: bool = not sys.platform.startswith('win')
_FUZE3_ALLOWED: bool = sys.platform.startswith('linux')


def _prepare_fuze():
    global _FUZE_READY
    assert _FUZE_ALLOWED, 'Windows is not supported'
    if _FUZE_READY: return
    Lib.import_cmd('fuse')
    Lib._ensure_lib_installed('fuse', 'fusepy')
    _FUZE_READY = True

def _add_proc(path: str, proc: threading.Thread):
    global _ALL_FUZES
    _ALL_FUZES[path] = proc

def _kill_proc(path: str, timeout: int = 5, force: bool = False):
    global _ALL_FUZES
    if _ALL_FUZES.get(path):
        proc = _ALL_FUZES.pop(path)
        proc.unmount(timeout)
    elif force:
        logger.warning(f'Forcefully unmounting {path}')
        umount = Lib.import_cmd('umount')
        umount(path).val

    

def _kill_all(fuzername: str = None, timeout: int = 5):
    """
    Kills all active mounts. If name is provided, will kill those matching the
    fusername, otherwise, removes all.
    """
    global _ALL_FUZES
    fuzerlist = list(_ALL_FUZES.keys()) if not fuzername else [k for k,v in _ALL_FUZES.items() if v.fuzer == fuzername]
    for path in fuzerlist:
        proc = _ALL_FUZES.pop(path)
        proc.unmount(timeout)


try: from lazy import CloudAuthz
except ImportError: CloudAuthz = object

_authz: CloudAuthz = None

def get_cloudauthz():
    global _authz
    if _authz is None:
        from lazy import CloudAuthz
        _authz = CloudAuthz
    return _authz

def run_fuze(fs: FuseSystemType, mount_point: MountPoint, foreground: bool = True, threads: bool = False, ready_file: bool = False, ops_class: 'FUSEr' = None):
    """Mount stuff in a local directory
    This uses fusepy to make it appear as if a given path on an fsspec
    instance is in fact resident within the local file-system.
    This requires that fusepy by installed, and that FUSE be available on
    the system (typically requiring a package to be installed with
    apt, yum, brew, etc.).
    Parameters
    ----------
    fs: FuseSystemType
        From one of the compatible implementations
    mount_point: MountPoint
        An empty directory on the local file-system where the contents of
        the remote path will appear.
    foreground: bool
        Whether or not calling this function will block. Operation will
        typically be more stable if True.
    threads: bool
        Whether or not to create threads when responding to file operations
        within the mounter directory. Operation will typically be more
        stable if False.
    ready_file: bool
        Whether the FUSE process is ready. The `.fuse_ready` file will
        exist in the `mount_point` directory if True. Debugging purpose.
    ops_class: FUSEr or Subclass of FUSEr
        To override the default behavior of FUSEr. For Example, logging
        to file.
    """
    from fsspec.fuse import FUSE, FUSEr
    if not ops_class: ops_class = FUSEr
    mount_point.target_path.mkdir(parents=True, exist_ok=True)
    func = lambda: FUSE(
        ops_class(fs, mount_point.source_path, ready_file=ready_file), 
        mount_point.target_path_str, 
        nothreads=not threads, 
        foreground=foreground)
    if not foreground:
        th = threading.Thread(target=func)
        th.daemon = True
        th.start()
        logger.info(f'Started Mount {mount_point.source} -> {mount_point.target_path_str} in Background')
        mount_point.thread = th
        _add_proc(mount_point.mount_point, mount_point)
        return th
    else:  # pragma: no cover
        try:
            logger.info(f'Started Mount {mount_point.source} -> {mount_point.target_path_str} in Foreground')
            func()
        except KeyboardInterrupt:
            _kill_proc(mount_point.mount_point)
            #sys.exit()



"""
Will use MemoryFileSystem as default
"""

class BaseFuzerCls:
    """
    Class that manages filesystem fuse mounts
    """

    _FSX: ModuleType = None
    _FSX_LIB: str = 'fsspec'
    _FSX_MODULE: Optional[str] = 'fsspec.implementations.memory'
    _FSX_CLS: str = 'MemoryFileSystem'
    _AUTHZ: 'CloudAuthz' = None

    @classmethod
    def _ensure_lib(cls, *args, **kwargs):
        if cls._FSX is not None: return
        cls._FSX = Lib.import_lib(cls._FSX_LIB)
        if cls._FSX_MODULE: cls._FSX = import_module(cls._FSX_MODULE, package=cls._FSX_LIB)

    @classmethod
    def get_authz(cls, reload: bool = False, **config):
        if cls._AUTHZ and not reload: return cls._AUTHZ
        cls._AUTHZ = get_cloudauthz()
        if config: cls._AUTHZ.update_authz(**config)
        return cls._AUTHZ

    @classmethod
    def get_configz(cls, reload: bool = False, **config):
        return {}

    @classmethod
    def get_filesystem(cls, fs_args: Dict[str, Any] = {}, *args, **kwargs) -> FuseSystemType:
        cls._ensure_lib()
        configz = cls.get_configz(*args, **kwargs)
        if fs_args: configz.update(fs_args)
        return getattr(cls._FSX, cls._FSX_CLS)(**configz)
    
    @classmethod
    def _strip_prefix(cls, source: str):
        uri_splits = source.split('://', maxsplit=1)
        if len(uri_splits) > 1: return uri_splits[-1].strip()
        return source

    @classmethod
    def mount(cls, source: str, mount_path: str, ready_file: bool = True, foreground: bool = False, threads: bool = False, use_mp: bool = False, daemon: bool = True, fs_args: Dict[str, Any] = {}, cleanup: bool = True, *args, **kwargs) -> None:
        """ 
        Mounts a source path/bucket/etc with this FuzeCls at mount_path

        source: Bucket/path that will be the source of the mount
        mount_path: the directory that will be mounted to
        foreground: return a thread blocking function, otherwise will be a background thread that will need to be killed explicitly.
        fs_args: Dict[str, Any] that will be passed when initializing the FileSystem (such as auth)
        cleanup: removes the mount_path on unmount
        """
        assert mount_path not in _ALL_FUZES, f'{mount_path} is already mounted from {source}'
        _prepare_fuze()
        logger.info(f'[{cls.__name__}] Starting Mount')
        fs = cls.get_filesystem(fs_args, *args, **kwargs)
        _mp = MountPoint(fuzer = cls.__name__, source=source, mount_point=mount_path, cleanup=cleanup)
        _mp.mount(fs, foreground = foreground, threads = threads, ready_file = ready_file, use_mp = use_mp, daemon = daemon)
        #run_fuze(fs, mp, foreground=foreground, threads=threads, ready_file=ready_file)
    
    @classmethod
    def unmount(cls, mount_path: str, timeout: int = 5, force: bool = False) -> None:
        """ 
        Unmounts an existing mount at mount_path
        """
        return _kill_proc(mount_path, timeout = timeout, force = force)
    
    @classmethod
    def unmount_all(cls, global_unmount: bool = False, timeout: int = 5) -> None:
        """ 
        Unmounts all existing mounts for this Fuze Class, i.e. all s3, or all gcs. 
        
        If global_unmount is True, then unmount all, regardless of class
        """
        fzername = None if global_unmount else cls.__name__
        return _kill_all(fzername, timeout = timeout)







from . import _imports
from ._imports import LazyLib

from . import _asyncify
from ._asyncify import asyncify

from . import pathio
from .pathio import Path, PathLike

from . import _envcls
from ._envcls import EnvCls


"""
Core Function based on 
https://github.com/kbairak/pipepy/blob/master/src/pipepy/pipepy.py

"""

from .base import *
from .exceptions import CmdError
from .utils import ExecFile, is_iterable

from typing import Dict

CMD_ALWAYS_RAISE = False
CMD_ALWAYS_STREAM = False
CMD_INTERACTIVE = False

_CMD_JOBS: Dict[int, "Cmd"] = {}


def jobs(): return list(_CMD_JOBS.values())


def wait_jobs(timeout=None):
    for command in jobs():
        command.wait(timeout)

def set_always_raise(value):
    global CMD_ALWAYS_RAISE
    CMD_ALWAYS_RAISE = value

def set_always_stream(value):
    global CMD_ALWAYS_STREAM
    CMD_ALWAYS_STREAM = value

def set_interactive(value):
    global CMD_INTERACTIVE
    CMD_INTERACTIVE = value


class Cmd:
    # Init and copies
    def __init__(self, *args, _lazy=False, _input=None, _stream_stdout=None,
                 _stream_stderr=None, _stream=None, _text=True,
                 _encoding="UTF-8", _raise=None, **kwargs):
        """ Initialize a Cmd object.
            Usually you will not need to call this directly, but invoke
            `__call__` to customize existing Cmd objects. Most arguments are
            not meant to be set by users directly, but by other functions or
            operators.
            The ones you may set yourself are:
            - args, kwargs: Customizes the command that is passed to the shell
            - _stream_stdout, _stream_stderr: Determines whether the relevant
                output stream of the command will be stored or whether it will
                be passed on to the relevant stream of the underlying Python
                process. `_stream` applies to both streams takes precedence
                over the other two. If you get a copy of the command with
                `.stream()`, the `_stream` parameter will be set. If not set,
                the `cmd.CMD_ALWAYS_STREAM` value will be respected (defaults to
                False)
            - _text, _encoding: Whether the `str` or the `bytes` type will be
                used for input and output. `_text` defaults to `True`. The
                associated subprocesses and files will be opened with this
                setting. Also, strings or byte sequences passed as input/output
                will be automatically converted to the correct type, using the
                encoding described by `_encoding` (defaults to `UTF-8`)
            - _raise: Whether a command will raise an exception if its
                returncode is not 0. If not set, the `cmd.ALWAYS_RAISE`
                setting will be respsected (defaults to `False`). If you don't
                set it, or if you set it to `False`, you can still raise an
                exception by calling `command.raise_for_returncode()` (similar
                to `request`'s `response.raise_for_status()`)
            The ones that are set by functions or operators are:
            - _lazy: Whether this instance will be evaluated again after having
                been evaluated once. Cmd objects created with the
                constructor will have this set to False but copies returned
                from `__call__` or other helper functions will have this set to
                True
            - _input: Where the command's input comes from. Will be populated
                  by pipe operations `"foo\nbar\n" | grep("foo")` will be
                  equivalent to`grep("foo", _input="foo\nbar\n")`)
            - _stream: Whether the output streams will be captured or passed on
                to the relevant streams of the underlying Python process. Will
                be set to True to the copy returned by `.stream()`
        """

        self._args = self._convert_args(args, kwargs)
        self._lazy = _lazy
        self._input = _input
        self._stream_stdout = _stream_stdout
        self._stream_stderr = _stream_stderr
        self._stream = _stream
        self._text = _text
        self._encoding = _encoding
        self._raise = _raise

        self._process = None
        self._input_consumed = False

        self._returncode = None
        self._stdout = None
        self._stderr = None

    def __call__(self, *args, _input=None, _stream_stdout=None,
                 _stream_stderr=None, _stream=None, _text=None, _encoding=None,
                 _raise=None, **kwargs):
        """ Make and return a copy of `self`, overriding some of its
            parameters.
            The copy will be lazy, ie if evaluated once and its output accessed
            a second time, the second time will return the stored values and
            not trigger another evaluation.
            If called without any arguments, will immediately trigger an
            evaluation.
        """

        force = (not args and
                 _input is None and
                 _stream_stdout is None and
                 _stream_stderr is None and
                 _stream is None and
                 _text is None and
                 _encoding is None and
                 _raise is None and
                 not kwargs)

        args = self._args + list(args)
        if _input is None: _input = self._input
        if _stream_stdout is None: _stream_stdout = self._stream_stdout
        if _stream_stderr is None: _stream_stderr = self._stream_stderr
        if _stream is None: _stream = self._stream
        if _text is None: _text = self._text
        if _encoding is None: _encoding = self._encoding
        if _raise is None: _raise = self._raise

        result = self.__class__(*args, _lazy=True, _input=_input,
                                _stream_stdout=_stream_stdout,
                                _stream_stderr=_stream_stderr, _stream=_stream,
                                _text=_text, _encoding=_encoding,
                                _raise=_raise, **kwargs)
        
        if force: result._evaluate()
        return result

    def __sub__(left, right):
        """ Alternate method of adding switches to commands:
                >>> ls - 'l'
                >>> # is equivalent to
                >>> ls('-l')
            If the right operand is longer than 1 character, 2 dashes will be
            used:
                >>> ls - 'escape'
                >>> # is equivalent to
                >>> ls('--escape')
        """

        if len(right) == 1: return left(f"-{right}")
        else: return left(f"--{right}")

    def __getattr__(self, attr):
        """ Alternate way of pasing arguments to commands. Essentially
                >>> git = Cmd('git')
                >>> git.status
            should be equivalent to
                >>> git('status')
        """

        return self.__class__(*(self._args + [attr]), _lazy=self._lazy,
                              _input=copy(self._input),
                              _stream_stdout=self._stream_stdout,
                              _stream_stderr=self._stream_stderr,
                              _stream=self._stream, _text=self._text,
                              _encoding=self._encoding, _raise=self._raise)

    def __copy__(self):
        return self.__class__(*self._args, _lazy=True,
                              _input=copy(self._input),
                              _stream_stdout=self._stream_stdout,
                              _stream_stderr=self._stderr,
                              _stream=self._stream, _text=self._text,
                              _encoding=self._encoding)

    @staticmethod
    def _convert_args(args, kwargs):
        """ Do some fancy processing of arguments. The intention is to enable
            things like:
                >>> Cmd('sleep', 10)
                >>> # Equivalent to
                >>> Cmd('sleep', '10')
                >>> Cmd('ls', I='foo')
                >>> # Equivalent to
                >>> Cmd('ls', '-I', 'foo')
                >>> Cmd('ls', sort="size")
                >>> # Equivalent to
                >>> Cmd('ls', '--sort=size')
                >>> Cmd('ls', escape=True)
                >>> # Equivalent to
                >>> Cmd('ls', '--escape')
            Because positional arguments come before keyword arguments and the
            order of keyword arguments is not guaranteed, you can apply
            multiple function calls to enforce your preferred ordering:
                >>> Cmd('ls', sort="size")('-l')
                >>> # Equivalent to
                >>> Cmd('ls', '--sort=size', '-l')
        """

        args = [str(arg) for arg in args]

        final_args = []
        for arg in args:
            arg = str(arg)
            globbed = glob(arg, recursive=True)
            if globbed: final_args.extend(globbed)
            else: final_args.append(arg)

        for key, value in kwargs.items():
            key = key.replace('_', '-')
            if value is True: final_args.append(f"--{key}")
            elif value is False: final_args.append(f"--no-{key}")
            elif len(key) == 1: final_args.extend([f"-{key}", value])
            else: final_args.append(f"--{key}={value}")
        
        return final_args

    # Lifetime implementation
    def _evaluate(self):
        """ Start an evaluations, Lazy commands that have been evaluated before
            will abort. The lifetime of a command being evaluated consists of 3
            steps:
            - Starting the process
            - Feeding the input to the process's stdin, if it has been
              configured to do so with pipes etc
            - Waiting for the command to finish and capture its result
            Piping operations usually don't call this method to evaluate but
            manually invoke the first 2 steps, utilizing the output of the
            process and then calling `wait`, which is a public method.
        """

        if self._returncode is not None and self._lazy: return

        self._start_background_job()
        self._feed_input()
        self.wait()

    def _start_background_job(self, stdin_to_pipe=False):
        """ Starts the process that will carry out the command. If the process
            has already been started, it will abort. If the input to this
            command is another Cmd object, its background process will be
            started too via this method (so it will recursively start all
            background processes of a pipe chain if necessary) and its stdout
            will be connected to our stdin.
        """

        if self._process is not None and self._lazy: return

        if isinstance(self._input, Cmd):
            if self._input._returncode is not None: stdin = PIPE
            else:
                self._input._start_background_job(stdin_to_pipe=stdin_to_pipe)
                stdin = self._input._process.stdout
        
        elif (is_iterable(self._input) or stdin_to_pipe or isinstance(self._input, ExecFile)): stdin = PIPE
        else: stdin = None

        if self._stream_stdout is None and self._stream is None: stdout = None if CMD_ALWAYS_STREAM else PIPE
        elif self._stream_stdout is None and self._stream is not None: stdout = None if self._stream else PIPE
        else: stdout = None if self._stream_stdout else PIPE

        if self._stream_stderr is None and self._stream is None: stderr = None if CMD_ALWAYS_STREAM else PIPE
        elif self._stream_stderr is None and self._stream is not None: stderr = None if self._stream else PIPE
        else: stderr = None if self._stream_stderr else PIPE

        self._process = Popen(self._args, stdin=stdin, stdout=stdout, stderr=stderr, text=self._text)
        _CMD_JOBS[self._process.pid] = self

    def _feed_input(self):
        """ If the command has been configured to receive special input via its
            `_input` parameter, ie via pipes or input redirects, the input will
            be passed to the command during this step.
        """

        if self._input_consumed and self._lazy: return

        left = self._input
        if isinstance(left, Cmd):
            if left._returncode is not None:
                chunk = left.stdout
                if self._text:
                    try: chunk = chunk.decode(self._encoding)
                    except AttributeError: pass
                else:
                    try: chunk = chunk.encode(self._encoding)
                    except AttributeError: pass
                self._process.stdin.write(chunk)
                self._process.stdin.flush()
                self._process.stdin.close()
            else:
                left._start_background_job()
                left._feed_input()
        
        elif isinstance(left, ExecFile):
            with left.path.open(mode="r" if self._text else "rb", encoding=self._encoding if self._text else None) as f:
                for line in f:
                    self._process.stdin.write(line)
                    self._process.stdin.flush()
                self._process.stdin.close()
        elif is_iterable(left):
            if isinstance(left, (str, bytes)): left = [left]
            
            for chunk in left:
                if self._text:
                    try: chunk = chunk.decode(self._encoding)
                    except AttributeError: pass
                else:
                    try: chunk = chunk.encode(self._encoding)
                    except AttributeError: pass
                self._process.stdin.write(chunk)
                self._process.stdin.flush()
            
            self._process.stdin.close()

        self._input_consumed = True

    # Control lifetime
    def delay(self):
        """ Create and return a copy of `self` and perform 2 out of 3 steps of
            its evaluation, ie don't wait for its result.
            You can then manually `.wait()` for this command to finish, or you
            can try to evaluate it (by accessing its output), which will cause
            its normal `_evaluate` method to run, which will skip the first 2
            steps and internally call `.wait()` before capturing its output
                >>> sleep = Cmd('sleep')
                >>> job = sleep(5).delay()
                >>> if job:  # This will wait for the command to finish
                ...     print("Job finished")
            You should take care to manually wait for background commands to
            finish yourself. If the python process ends, all its child
            processes will end too and your command may be killed abrubtly.
        """

        result = copy(self)
        result._start_background_job()
        result._feed_input()
        return result

    def wait(self, timeout=None):
        """ Wait for a process to finish and store the result.
            This is called internally by pipe operations, but can also be
            called by the user for a background command that has been created
            with `.delay()`.
                >>> sleep = Cmd('sleep')
                >>> job = sleep(5).delay()
                >>> job.wait()
                >>> print("Job finished")
        """

        try:
            self._stdout, self._stderr = self._process.\
                communicate(timeout=timeout)
            self._returncode = self._process.wait()
        except TimeoutExpired: raise
        except Exception:
            if self._process.stdout is not None: self._stdout = self._process.stdout.read()
            else: self._stdout = "" if self._text else b""
            if self._process.stderr is not None: self._stderr = self._process.stderr.read()
            else: self._stderr = "" if self._text else b""
            self._returncode = self._process.wait(timeout)
        
        try: del _CMD_JOBS[self._process.pid]
        except KeyError: pass

        job = self
        while isinstance(job._input, Cmd):
            job = job._input
            job.wait(timeout)

        raise_exception = self._raise
        if raise_exception is None: raise_exception = CMD_ALWAYS_RAISE
        if raise_exception: self.raise_for_returncode()

    def raise_for_returncode(self):
        """ Raise an exception if the command's returncode is not 0.
            Will be called automatically for all commands that are not created
            with `.quiet` if `cmd.ALWAYS_RAISE` is True.
            The exception will have the `returncode`, `stdout` and `stderr`
            properties.
        """

        if self.returncode != 0: raise CmdError(self._returncode, self._stdout, self._stderr)

    # Getting output
    @property
    def returncode(self):
        """ Evaluate the command and return its returncode. """

        self._evaluate()
        return self._returncode

    @property
    def stdout(self):
        """ Evaluate the command and return its stdout. """

        self._evaluate()
        return self._stdout

    @property
    def stderr(self):
        """ Evaluate the command and return its stderr. """

        self._evaluate()
        return self._stderr
    
    @property
    def val(self):
        """ Evaluate the command and tries to parse the result into the type class. """
        _val = self.stdout or self.stderr
        # Assumes its json
        if (_val.startswith('{') and _val.endswith('}') or _val.startswith('[') and _val.endswith(']')):
            try: return OrJson.loads(_val)
            except: pass
        # assumes article or document
        if _val.count('\n\n') > 0: return _val
        # assumes list
        if _val.count('\n') >= 2: return [i for i in _val.split('\n') if i]
        return _val.strip()

    def __str__(self):
        """ Return stdout as string, even if the command has `_text=False`. """

        try: return self.stdout.decode(self._encoding)
        except UnicodeDecodeError: return str(self.stdout)
        except AttributeError: return self.stdout

    def __bool__(self):
        """ Use in boolean expressions.
            Usage:
                >>> git = Cmd('git')
                >>> grep = Cmd('grep')
                >>> if git('branch') | grep('my_feature'):
                ...     print("Branch found")
        """

        return self.returncode == 0

    def __iter__(self):
        """ Support the iteration interface:
            Usage:
                >>> ls = Cmd('ls')
                >>> for name in ls:
                ...     print(ls.upper())
        """

        if self._stdout is not None:
            yield from self._stdout.splitlines()
        else:
            self._start_background_job()
            self._feed_input()
            yield from self._process.stdout
            self.wait()

    def iter_words(self):
        """ Iterate over the *words* of the output of the command.
                >>> ps = Cmd('ps')
                >>> list(ps.iter_words())
                <<< ['PID', 'TTY', 'TIME', 'CMD',
                ...  '11439', 'pts/5', '00:00:00', 'zsh',
                ...  '15532', 'pts/5', '00:00:10', 'ptipython',
                ...  '15539', 'pts/5', '00:00:00', 'ps']
        """

        for line in self:
            yield from line.split()

    def as_table(self):
        """ Usage:
            >>> ps = Cmd('ps')
            >>> ps.as_table()
            <<< [{'PID': '11233', 'TTY': 'pts/4', 'TIME': '00:00:01',
            ...   'CMD': 'zsh'},
            ...  {'PID': '17673', 'TTY': 'pts/4', 'TIME': '00:00:08',
            ...   'CMD': 'ptipython'},
            ...  {'PID': '18281', 'TTY': 'pts/4', 'TIME': '00:00:00',
            ...   'CMD': 'ps'}]
        """
        lines = self.stdout.splitlines()
        fields = lines[0].split()
        result = []
        for line in lines[1:]:
            item = {}
            for i, value in enumerate(line.split(maxsplit=len(fields) - 1)):
                item[fields[i]] = value
            result.append(item)
        return result

    def __repr__(self):
        """ Return some useful information about the Cmd object.
            If `cmd.INTERACTIVE` is set, it will evaluate and print the
            command's output, offering a semblance of an interactive shell.
                >>> ls = Cmd('ls')
                >>> ls
                <<< Cmd('ls')
                >>> cmd.INTERACTIVE = True
                >>> ls
                <<< main.py files.txt
        """

        if CMD_INTERACTIVE: return self._interactive_repr()
        else: return self._normal_repr()

    def _normal_repr(self):
        result = [self.__class__.__name__, "("]
        result.append(', '.join((repr(arg) for arg in self._args)))
        if self._input is not None:
            result.append(f", _input={self._input!r}")
        if self._returncode is not None:
            result.append(f", _returncode={self._returncode}")
        if self._stdout:
            result.append(f", _stdout={reprlib.repr(self._stdout)}")
        if self._stderr:
            result.append(f", _stderr={reprlib.repr(self._stderr)}")
        result.append(")")
        return ''.join(result)

    def _interactive_repr(self):
        return self.stdout + self.stderr

    # Redirect output
    def __gt__(left, right):
        """ Write output to file or file-like object
            Usage:
                >>> ps = Cmd('ps')
                >>> ps > 'progs.txt'
        """

        if isinstance(right, (Path, pathlib.Path, str)):
            right: PathLike = Path.get_path(right, resolve=True)
            with right.open("w" if left._text else "wb", encoding=left._encoding if left._text else None) as f:
                if left._returncode is None:
                    for line in left:
                        f.write(line)
                
                else: f.write(left.stdout)
                    
        elif isinstance(right, io.IOBase):
            right.seek(0)
            right.truncate()
            if left._returncode is None:
                for line in left:
                    right.write(line)
            
            else: right.write(left.stdout)
        else: return NotImplemented

    def __rshift__(left, right):
        """ Append output to file or file-like object
            Usage:
                >>> ps = Cmd('ps')
                >>> ps >> 'progs.txt'
        """
        if isinstance(right, (Path, pathlib.Path, str)):
            right: PathLike = Path.get_path(right, resolve=True)
            with right.open("a" if left._text else "ab", encoding=left._encoding if left._text else None) as f:
                if left._returncode is None:
                    for line in left:
                        f.write(line)
                
                else: f.write(left.stdout)
        
        elif isinstance(right, io.IOBase):
            right.read()  # Move pointer to end
            if left._returncode is None:
                for line in left:
                    right.write(line)
            else: right.write(left.stdout)
        
        else: return NotImplemented

    def __lt__(left, right):
        """ Read input from file or file-like object
            Usage:
                >>> grep = Cmd('grep')
                >>> grep('python') < 'progs/txt'
        """

        if isinstance(right, (Path, pathlib.Path, str)): return left(_input=ExecFile(right))
        elif isinstance(right, io.IOBase): return left(_input=iter(right))
        else: return NotImplemented

    # Pipes
    def __or__(left, right):
        return Cmd._pipe(left, right)

    def __ror__(right, left):
        return Cmd._pipe(left, right)

    @staticmethod
    def _pipe(left, right):
        """ Support pipe operations. Depending on the operands, slightly
            different behaviors emerge:
            1. If both operands are Cmd objects, the returned object will,
               upon evaluation, start both operands as background processes and
               connect the left's stdout to the right's stdin
                >>> ls = Cmd('ls')
                >>> grep = Cmd('grep')
                >>> print(ls | grep('files'))
                <<< files.txt
            2. If only the right operand is a Cmd object, the left operand
               will be used as input. The left operand can be any iterable
               object, including lists, strings or generators
                >>> grep = Cmd('grep')
                >>> print(["foo\n", "bar\n"] | grep("foo"))
                <<< foo
                >>> def my_input():
                ...     yield "foo\n"
                ...     yield "bar\n"
                >>> print(my_input() | grep("foo"))
                <<< foo
            3. If only the the left operand is a Cmd object and the right
               object is a function, then the command will be evaluated and its
               output will be passed as arguments to the function:
               - If the function's arguments are a subset of [returncode,
                 output, errors], the command will be waited and its output
                 will be passed at once to the function
               - If the function's arguments are a subset of [stdout, stderr],
                 the command will be run in the background and its stdout and
                 stderr streams will be made available to the function
               The ordering of the arguments doesn't matter since the
               function's signature will be inspected to determine the
               appropriate behavior
            4. If only the the left operand is a Cmd object and the right
               object is a generator (the return value of a function that
               `yield`s), then the generator will receive non-empty lines from
               the command's output and the return value of the pipe operation
               will be another generator that will yield whatever the original
               generator yielded
                >>> def my_generator():
                ...     line = yield
                ...     while True:
                ...         line = (yield line.upper())
                >>> list("aaa\nbbb" | cat | my_generator())
                <<< ["AAA", "BBB"]
        """

        if isinstance(left, Cmd) and isinstance(right, Cmd): return right(_input=left)
        elif isinstance(right, Cmd):
            if is_iterable(left): return right(_input=left)
            else: return NotImplemented
        
        elif isinstance(left, Cmd):
            if callable(right): return left._send_output_to_function(right)
            elif isinstance(right, types.GeneratorType): return left._send_output_to_generator(right)
            else: return NotImplemented
        
        else: return NotImplemented

    # Help with pipes
    def _send_output_to_function(self, func):
        """ Implement the "pipe to function" functionality """

        error = TypeError(f"Cannot pipe to {func!r}: Invalid function signature")
        parameters = inspect.signature(func).parameters
        if not parameters: raise error
        if not all((value.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD for value in parameters.values())):
            raise error
        
        keys = set(parameters.keys())
        if keys <= {'returncode', 'output', 'errors'}:
            arguments = {'returncode': self.returncode,
                         'output': self.stdout,
                         'errors': self.stderr}
            kwargs = {key: value for key, value in arguments.items() if key in keys}
            return func(**kwargs)
        
        elif keys <= {'stdout', 'stderr'}:
            self._start_background_job()
            self._feed_input()
            arguments = {'stdout': self._process.stdout,
                         'stderr': self._process.stderr}
            
            kwargs = {key: value for key, value in arguments.items() if key in keys}
            result = func(**kwargs)
            if isinstance(result, types.GeneratorType):
                # Make returned generator wait for the current command on exit
                def generator():
                    yield from result
                    self.wait()
                return generator()
            else:
                self.wait()
                return result
        
        else: raise error

    def _send_output_to_generator(self, generator):
        """ Implement the "pipe to generator" functionality """

        def result():
            self._start_background_job()
            self._feed_input()
            stdout = (line.strip() + "\n" for line in self._process.stdout if line.strip())
            
            try: next_input = next(generator)
            except StopIteration:
                generator.close()
                return
            while True:
                if next_input is not None:
                    yield next_input
                try: next_input = generator.send(next(stdout))
                except StopIteration: break
            generator.close()
        
        return result()

    # `with` statements
    def __enter__(self):
        """ Start a job in the background and allow the code block to interact
            with *both* its input and output:
                >>> grep = Cmd('grep')
                >>> with grep("foo") as (stdin, stdout, stderr):
                ...     stdin.write("foo\n")
                ...     stdin.write("bar\n")
                ...     stdin.close()
                ...     print(stdout.read())
                <<< foo
        """

        self._start_background_job(stdin_to_pipe=True)

        job = self
        while isinstance(job._input, Cmd):
            job = job._input

        return job._process.stdin, self._process.stdout, self._process.stderr

    def __exit__(self, exc_type, exc_val, exc_tb):
        job = self
        while isinstance(job._input, Cmd):
            job = job._input
        job._process.stdin.close()

        self.wait()
        job = self
        while isinstance(job._input, Cmd):
            job = job._input
            job.wait()

    # Forward calls to background process
    def _map_to_background_process(method):
        """ Expose the `send_signal`, `terminate` and `kill` methods of Popen
            objects to Cmd objects.
        """

        def func(self, *args, **kwargs):
            if self._process is None: raise TypeError(f"Cannot call '{method}' on non-background process")
            getattr(self._process, method)(*args, **kwargs)
        
        return func

    send_signal = _map_to_background_process('send_signal')
    terminate = _map_to_background_process('terminate')
    kill = _map_to_background_process('kill')


import re
from pydantic import Field
from lazy.types.base import *
from lazy.types.pyd import *

from .core import ConfigCls
from .types import *

__all__ = (
    'DBBaseConfigCls',
    'PostgresConfigz',
    'MysqlConfigz',
    'RedisConfigz',
    'ElasticsearchConfigz',
    'SMTPEmailBaseConfigCls',
    'EmailBaseConfigCls',
    'AppBaseConfigCls',
    'FastAPIBaseConfigCls',
)


"""
Database Configs
"""
class DBBaseConfigCls(ConfigCls):
    host: Optional[str] = "127.0.0.1"
    port: Optional[int] = 5432
    username: Optional[str] = None
    password: Optional[str] = None
    config: Optional[DictStr] = None
    database: Optional[str] = None

class PostgresConfigz(DBBaseConfigCls):
    username: Optional[str] = 'postgres'
    database: Optional[str] = 'postgres'

class MysqlConfigz(DBBaseConfigCls):
    port: Optional[int] = 3306
    username: Optional[str] = 'mysql'
    database: Optional[str] = 'mysql'
    
class RedisConfigz(DBBaseConfigCls):
    port: Optional[int] = 6379
    database: Optional[int] = 0

class ElasticsearchConfigz(DBBaseConfigCls):
    port: Optional[int] = 9200


"""
Commonly Used App Configs

To use for your specific app, you can subclass AppBaseConfigCls

MyAppConfigz(AppBaseConfigCls): pass

all env variables would be translated be prefixed with MYAPP_
title       -> MYAPP_TITLE
description -> MYAPP_DESCRIPTION

"""
class SMTPEmailBaseConfigCls(ConfigCls):
    tls: Optional[bool] = True
    port: Optional[int] = 587
    host: Optional[str] = 'smtp.gmail.com'
    user: Optional[str] = 'user'
    password: Optional[str] = 'password'

    class Config:
        env_prefix = "SMTP_"

class EmailBaseConfigCls(ConfigCls):
    from_email: Optional[str] = None
    from_name: Optional[str] = None
    reset_token_expire_hrs: Optional[int] = 48
    templates_dir: Optional[PathStr] = './email-templates'
    enabled: Optional[bool] = False
    smtp_config: Optional[SMTPEmailBaseConfigCls] = Field(default_factory = SMTPEmailBaseConfigCls)

    class Config:
        env_prefix = "EMAILS_"


class AppBaseConfigCls(ConfigCls):
    title: Optional[str] = 'Lazy App'
    project: Optional[str] = 'Lazy Project'
    description: Optional[str] = 'Just a super secret app'
    version: Optional[str] = '0.0.1-alpha'
    domain: Optional[str] = None
    host: Optional[str] = '0.0.0.0'
    port: Optional[int] = 8080
    log_level: Optional[str] = 'info'
    config_dir: Optional[PathStr] = None

    author: Optional[str] = None
    maintainer: Optional[str] = None

    ### Auth Configs ###
    authuser: Optional[str] = None
    authpass: Optional[str] = None
    apikey: Optional[str] = None
    access_token: Optional[str] = None
    token: Optional[str] = None
    jwt_token: Optional[str] = None
    token_expire_mins: Optional[int] = 60 * 24 # 60 minutes * 24 hours

    authorized_users: Optional[ListStr] = []
    allowed_origins: Optional[ListStr] = []
    cors_origins: Optional[ListStr] = []

    open_registration: Optional[bool] = False

    ### Initialization ###
    first_superuser: Optional[str] = None
    first_superuser_role: Optional[str] = None
    first_superuser_pw: Optional[str] = None # Don't use this unless necessary
    first_superuser_pw_b64: Optional[Base64] = None
    first_superuser_pw_bgz: Optional[Base64Gzip] = None
    init_seed_data: Optional[Json] = {} # Json format
    init_seed_data_b64: Optional[JsonB64Str] = {} # Json in Base64 format
    init_seed_data_bgz: Optional[JsonB64GZipStr] = {} # Json in Base64 GZIP format    

    class Config:
        env_prefix = "APP_"

_path_prefix_pattern = re.compile(r'(/[\w.-]+)+')

try: from fastapi import FastAPI
except: FastAPI = object

class FastAPIBaseConfigCls(ConfigCls):
    root_path: Optional[str] = None
    api_prefix: Optional[str] = '/api/v1'
    xapi_prefix: Optional[str] = '/xapi'
    graphql_prefix: Optional[str] = '/graphql'
    openapi_prefix: Optional[str] = ''
    openapi_url: Optional[str] = '/openapi.json'
    docs_url: Optional[str] = '/docs'
    redoc_url: Optional[str] = '/redoc'
    version: Optional[str] = '0.0.1'
    root_path_in_servers: Optional[bool] = False
    allow_credentials: Optional[bool] = True
    allowed_hosts: Optional[ListStr] = []
    allowed_origins: Optional[ListStr] = ["*"]
    allowed_methods: Optional[ListStr] = ["*"]
    allowed_headers: Optional[ListStr] = ["*"]

    app_config: Optional[AppBaseConfigCls] = Field(default_factory = AppBaseConfigCls)

    class Config:
        env_prefix = "FASTAPI_"
    
    @property
    def app_prefix(self) -> str:
        """
        root_path takes precendence over api_prefix
        """
        if self.root_path: return self.root_path
        return self.api_prefix

    @property
    def openapi_path(self):
        """ 
        Prefixes openapi_url with openapi_prefix/app_prefix 
        with openapi_prefix taking precedence
        """
        p: str = (self.openapi_prefix or self.app_prefix) + self.openapi_url
        ## Fix any double slashes when combining path
        p = p.replace('//', '/')
        return p
    
    def get_openapi_prefix(self):
        """ 
        since openapi_url should either be a fixed path, or have prefix
        if openapi_prefix is present, this should be None
        """
        if self.openapi_prefix in self.openapi_path: return ''
        return self.openapi_prefix

    def get_fastapi_config(self, app_config: Dict[str, Any] = None) -> Dict[str, Any]:
        """
        Returns Dict for FastAPI(**config)
        """
        if app_config: self.app_config.update_config(**app_config) # self.app_config = self.app_config.update_config(**app_config)
        return {
            'title': self.app_config.title,
            'version': self.app_config.version or self.version,
            'root_path': self.app_prefix,
            'openapi_prefix': self.get_openapi_prefix(),
            'openapi_url': self.openapi_path,
            'root_path_in_servers': self.root_path_in_servers,
        }
    
    def get_fastapi_app(self, app_config: Dict[str, Any] = None, **kwargs) -> 'FastAPI':
        """
        Initializes a new FastAPI App using Config Settings

        returns fastapi.FastAPI
        """
        from lazy import Lib
        fastapi = Lib.import_lib('fastapi')
        fastapi_cls = getattr(fastapi, 'FastAPI')
        config = self.get_fastapi_config(app_config)
        if kwargs: config.update(kwargs)
        return fastapi_cls(**config)
    
    def update_fastapi_middleware(self, app: 'FastAPI', auth_config: Dict[str, Any] = None):
        """
        Updates an existing FastAPI instance with middlewares
        Assumes fastapi is installed since it would be required to pass the app

        returns the same app
        """
        if auth_config: self.update_config(**auth_config) #self = self.update_config(**auth_config)
        from starlette.middleware.cors import CORSMiddleware
        app.add_middleware(
            CORSMiddleware,
            allow_origins=self.allowed_origins,
            allow_credentials=self.allow_credentials,
            allow_methods=self.allowed_methods,
            allow_headers=self.allowed_headers,
        )
        if self.allowed_hosts:
            from fastapi.middleware.trustedhost import TrustedHostMiddleware
            app.add_middleware(TrustedHostMiddleware, allowed_hosts=self.allowed_hosts)
        return app



class FuzeConfigz(ConfigCls):
    cache_dir: PathStr = '~/.cachez'

    






""" 
Source: https://github.com/grantjenks/python-diskcache/blob/71db7753610bcd4bd7adda28abeb2b0fdebbc31f/diskcache/persistent.py
"""

import operator as op
from collections import OrderedDict
from collections.abc import (
    ItemsView,
    KeysView,
    MutableMapping,
    Sequence,
    ValuesView,
)
from contextlib import contextmanager
from shutil import rmtree

from .core import ENOVAL, Cache


def _make_compare(seq_op, doc):
    "Make compare method with Sequence semantics."

    def compare(self, that):
        "Compare method for deque and sequence."
        if not isinstance(that, Sequence):
            return NotImplemented

        len_self = len(self)
        len_that = len(that)

        if len_self != len_that:
            if seq_op is op.eq:
                return False
            if seq_op is op.ne:
                return True

        for alpha, beta in zip(self, that):
            if alpha != beta:
                return seq_op(alpha, beta)

        return seq_op(len_self, len_that)

    compare.__name__ = '__{0}__'.format(seq_op.__name__)
    doc_str = 'Return True if and only if deque is {0} `that`.'
    compare.__doc__ = doc_str.format(doc)

    return compare


class Deque(Sequence):
    """Persistent sequence with double-ended queue semantics.
    Double-ended queue is an ordered collection with optimized access at its
    endpoints.
    Items are serialized to disk. Deque may be initialized from directory path
    where items are stored.
    >>> deque = Deque()
    >>> deque += range(5)
    >>> list(deque)
    [0, 1, 2, 3, 4]
    >>> for value in range(5):
    ...     deque.appendleft(-value)
    >>> len(deque)
    10
    >>> list(deque)
    [-4, -3, -2, -1, 0, 0, 1, 2, 3, 4]
    >>> deque.pop()
    4
    >>> deque.popleft()
    -4
    >>> deque.reverse()
    >>> list(deque)
    [3, 2, 1, 0, 0, -1, -2, -3]
    """

    def __init__(self, iterable=(), directory=None):
        """Initialize deque instance.
        If directory is None then temporary directory created. The directory
        will *not* be automatically removed.
        :param iterable: iterable of items to append to deque
        :param directory: deque directory (default None)
        """
        self._cache = Cache(directory, eviction_policy='none')
        self.extend(iterable)

    @classmethod
    def fromcache(cls, cache, iterable=()):
        """Initialize deque using `cache`.
        >>> cache = Cache()
        >>> deque = Deque.fromcache(cache, [5, 6, 7, 8])
        >>> deque.cache is cache
        True
        >>> len(deque)
        4
        >>> 7 in deque
        True
        >>> deque.popleft()
        5
        :param Cache cache: cache to use
        :param iterable: iterable of items
        :return: initialized Deque
        """
        # pylint: disable=no-member,protected-access
        self = cls.__new__(cls)
        self._cache = cache
        self.extend(iterable)
        return self

    @property
    def cache(self):
        "Cache used by deque."
        return self._cache

    @property
    def directory(self):
        "Directory path where deque is stored."
        return self._cache.directory

    def _index(self, index, func):
        len_self = len(self)

        if index >= 0:
            if index >= len_self:
                raise IndexError('deque index out of range')

            for key in self._cache.iterkeys():
                if index == 0:
                    try:
                        return func(key)
                    except KeyError:
                        continue
                index -= 1
        else:
            if index < -len_self:
                raise IndexError('deque index out of range')

            index += 1

            for key in self._cache.iterkeys(reverse=True):
                if index == 0:
                    try:
                        return func(key)
                    except KeyError:
                        continue
                index += 1

        raise IndexError('deque index out of range')

    def __getitem__(self, index):
        """deque.__getitem__(index) <==> deque[index]
        Return corresponding item for `index` in deque.
        See also `Deque.peekleft` and `Deque.peek` for indexing deque at index
        ``0`` or ``-1``.
        >>> deque = Deque()
        >>> deque.extend('abcde')
        >>> deque[1]
        'b'
        >>> deque[-2]
        'd'
        :param int index: index of item
        :return: corresponding item
        :raises IndexError: if index out of range
        """
        return self._index(index, self._cache.__getitem__)

    def __setitem__(self, index, value):
        """deque.__setitem__(index, value) <==> deque[index] = value
        Store `value` in deque at `index`.
        >>> deque = Deque()
        >>> deque.extend([None] * 3)
        >>> deque[0] = 'a'
        >>> deque[1] = 'b'
        >>> deque[-1] = 'c'
        >>> ''.join(deque)
        'abc'
        :param int index: index of value
        :param value: value to store
        :raises IndexError: if index out of range
        """

        def _set_value(key):
            return self._cache.__setitem__(key, value)

        self._index(index, _set_value)

    def __delitem__(self, index):
        """deque.__delitem__(index) <==> del deque[index]
        Delete item in deque at `index`.
        >>> deque = Deque()
        >>> deque.extend([None] * 3)
        >>> del deque[0]
        >>> del deque[1]
        >>> del deque[-1]
        >>> len(deque)
        0
        :param int index: index of item
        :raises IndexError: if index out of range
        """
        self._index(index, self._cache.__delitem__)

    def __repr__(self):
        """deque.__repr__() <==> repr(deque)
        Return string with printable representation of deque.
        """
        name = type(self).__name__
        return '{0}(directory={1!r})'.format(name, self.directory)

    __eq__ = _make_compare(op.eq, 'equal to')
    __ne__ = _make_compare(op.ne, 'not equal to')
    __lt__ = _make_compare(op.lt, 'less than')
    __gt__ = _make_compare(op.gt, 'greater than')
    __le__ = _make_compare(op.le, 'less than or equal to')
    __ge__ = _make_compare(op.ge, 'greater than or equal to')

    def __iadd__(self, iterable):
        """deque.__iadd__(iterable) <==> deque += iterable
        Extend back side of deque with items from iterable.
        :param iterable: iterable of items to append to deque
        :return: deque with added items
        """
        self.extend(iterable)
        return self

    def __iter__(self):
        """deque.__iter__() <==> iter(deque)
        Return iterator of deque from front to back.
        """
        _cache = self._cache

        for key in _cache.iterkeys():
            try:
                yield _cache[key]
            except KeyError:
                pass

    def __len__(self):
        """deque.__len__() <==> len(deque)
        Return length of deque.
        """
        return len(self._cache)

    def __reversed__(self):
        """deque.__reversed__() <==> reversed(deque)
        Return iterator of deque from back to front.
        >>> deque = Deque()
        >>> deque.extend('abcd')
        >>> iterator = reversed(deque)
        >>> next(iterator)
        'd'
        >>> list(iterator)
        ['c', 'b', 'a']
        """
        _cache = self._cache

        for key in _cache.iterkeys(reverse=True):
            try:
                yield _cache[key]
            except KeyError:
                pass

    def __getstate__(self):
        return self.directory

    def __setstate__(self, state):
        self.__init__(directory=state)

    def append(self, value):
        """Add `value` to back of deque.
        >>> deque = Deque()
        >>> deque.append('a')
        >>> deque.append('b')
        >>> deque.append('c')
        >>> list(deque)
        ['a', 'b', 'c']
        :param value: value to add to back of deque
        """
        self._cache.push(value, retry=True)

    def appendleft(self, value):
        """Add `value` to front of deque.
        >>> deque = Deque()
        >>> deque.appendleft('a')
        >>> deque.appendleft('b')
        >>> deque.appendleft('c')
        >>> list(deque)
        ['c', 'b', 'a']
        :param value: value to add to front of deque
        """
        self._cache.push(value, side='front', retry=True)

    def clear(self):
        """Remove all elements from deque.
        >>> deque = Deque('abc')
        >>> len(deque)
        3
        >>> deque.clear()
        >>> list(deque)
        []
        """
        self._cache.clear(retry=True)

    def count(self, value):
        """Return number of occurrences of `value` in deque.
        >>> deque = Deque()
        >>> deque += [num for num in range(1, 5) for _ in range(num)]
        >>> deque.count(0)
        0
        >>> deque.count(1)
        1
        >>> deque.count(4)
        4
        :param value: value to count in deque
        :return: count of items equal to value in deque
        """
        return sum(1 for item in self if value == item)

    def extend(self, iterable):
        """Extend back side of deque with values from `iterable`.
        :param iterable: iterable of values
        """
        for value in iterable:
            self.append(value)

    def extendleft(self, iterable):
        """Extend front side of deque with value from `iterable`.
        >>> deque = Deque()
        >>> deque.extendleft('abc')
        >>> list(deque)
        ['c', 'b', 'a']
        :param iterable: iterable of values
        """
        for value in iterable:
            self.appendleft(value)

    def peek(self):
        """Peek at value at back of deque.
        Faster than indexing deque at -1.
        If deque is empty then raise IndexError.
        >>> deque = Deque()
        >>> deque.peek()
        Traceback (most recent call last):
            ...
        IndexError: peek from an empty deque
        >>> deque += 'abc'
        >>> deque.peek()
        'c'
        :return: value at back of deque
        :raises IndexError: if deque is empty
        """
        default = None, ENOVAL
        _, value = self._cache.peek(default=default, side='back', retry=True)
        if value is ENOVAL:
            raise IndexError('peek from an empty deque')
        return value

    def peekleft(self):
        """Peek at value at front of deque.
        Faster than indexing deque at 0.
        If deque is empty then raise IndexError.
        >>> deque = Deque()
        >>> deque.peekleft()
        Traceback (most recent call last):
            ...
        IndexError: peek from an empty deque
        >>> deque += 'abc'
        >>> deque.peekleft()
        'a'
        :return: value at front of deque
        :raises IndexError: if deque is empty
        """
        default = None, ENOVAL
        _, value = self._cache.peek(default=default, side='front', retry=True)
        if value is ENOVAL:
            raise IndexError('peek from an empty deque')
        return value

    def pop(self):
        """Remove and return value at back of deque.
        If deque is empty then raise IndexError.
        >>> deque = Deque()
        >>> deque += 'ab'
        >>> deque.pop()
        'b'
        >>> deque.pop()
        'a'
        >>> deque.pop()
        Traceback (most recent call last):
            ...
        IndexError: pop from an empty deque
        :return: value at back of deque
        :raises IndexError: if deque is empty
        """
        default = None, ENOVAL
        _, value = self._cache.pull(default=default, side='back', retry=True)
        if value is ENOVAL:
            raise IndexError('pop from an empty deque')
        return value

    def popleft(self):
        """Remove and return value at front of deque.
        >>> deque = Deque()
        >>> deque += 'ab'
        >>> deque.popleft()
        'a'
        >>> deque.popleft()
        'b'
        >>> deque.popleft()
        Traceback (most recent call last):
            ...
        IndexError: pop from an empty deque
        :return: value at front of deque
        :raises IndexError: if deque is empty
        """
        default = None, ENOVAL
        _, value = self._cache.pull(default=default, retry=True)
        if value is ENOVAL:
            raise IndexError('pop from an empty deque')
        return value

    def remove(self, value):
        """Remove first occurrence of `value` in deque.
        >>> deque = Deque()
        >>> deque += 'aab'
        >>> deque.remove('a')
        >>> list(deque)
        ['a', 'b']
        >>> deque.remove('b')
        >>> list(deque)
        ['a']
        >>> deque.remove('c')
        Traceback (most recent call last):
            ...
        ValueError: deque.remove(value): value not in deque
        :param value: value to remove
        :raises ValueError: if value not in deque
        """
        _cache = self._cache

        for key in _cache.iterkeys():
            try:
                item = _cache[key]
            except KeyError:
                continue
            else:
                if value == item:
                    try:
                        del _cache[key]
                    except KeyError:
                        continue
                    return

        raise ValueError('deque.remove(value): value not in deque')

    def reverse(self):
        """Reverse deque in place.
        >>> deque = Deque()
        >>> deque += 'abc'
        >>> deque.reverse()
        >>> list(deque)
        ['c', 'b', 'a']
        """
        # GrantJ 2019-03-22 Consider using an algorithm that swaps the values
        # at two keys. Like self._cache.swap(key1, key2, retry=True) The swap
        # method would exchange the values at two given keys. Then, using a
        # forward iterator and a reverse iterator, the reversis method could
        # avoid making copies of the values.
        temp = Deque(iterable=reversed(self))
        self.clear()
        self.extend(temp)
        directory = temp.directory
        del temp
        rmtree(directory)

    def rotate(self, steps=1):
        """Rotate deque right by `steps`.
        If steps is negative then rotate left.
        >>> deque = Deque()
        >>> deque += range(5)
        >>> deque.rotate(2)
        >>> list(deque)
        [3, 4, 0, 1, 2]
        >>> deque.rotate(-1)
        >>> list(deque)
        [4, 0, 1, 2, 3]
        :param int steps: number of steps to rotate (default 1)
        """
        if not isinstance(steps, int):
            type_name = type(steps).__name__
            raise TypeError('integer argument expected, got %s' % type_name)

        len_self = len(self)

        if not len_self:
            return

        if steps >= 0:
            steps %= len_self

            for _ in range(steps):
                try:
                    value = self.pop()
                except IndexError:
                    return
                else:
                    self.appendleft(value)
        else:
            steps *= -1
            steps %= len_self

            for _ in range(steps):
                try:
                    value = self.popleft()
                except IndexError:
                    return
                else:
                    self.append(value)

    __hash__ = None  # type: ignore

    @contextmanager
    def transact(self):
        """Context manager to perform a transaction by locking the deque.
        While the deque is locked, no other write operation is permitted.
        Transactions should therefore be as short as possible. Read and write
        operations performed in a transaction are atomic. Read operations may
        occur concurrent to a transaction.
        Transactions may be nested and may not be shared between threads.
        >>> from diskcache import Deque
        >>> deque = Deque()
        >>> deque += range(5)
        >>> with deque.transact():  # Atomically rotate elements.
        ...     value = deque.pop()
        ...     deque.appendleft(value)
        >>> list(deque)
        [4, 0, 1, 2, 3]
        :return: context manager for use in `with` statement
        """
        with self._cache.transact(retry=True):
            yield


class Index(MutableMapping):
    """Persistent mutable mapping with insertion order iteration.
    Items are serialized to disk. Index may be initialized from directory path
    where items are stored.
    Hashing protocol is not used. Keys are looked up by their serialized
    format. See ``diskcache.Disk`` for details.
    >>> index = Index()
    >>> index.update([('a', 1), ('b', 2), ('c', 3)])
    >>> index['a']
    1
    >>> list(index)
    ['a', 'b', 'c']
    >>> len(index)
    3
    >>> del index['b']
    >>> index.popitem()
    ('c', 3)
    """

    def __init__(self, *args, **kwargs):
        """Initialize index in directory and update items.
        Optional first argument may be string specifying directory where items
        are stored. When None or not given, temporary directory is created.
        >>> index = Index({'a': 1, 'b': 2, 'c': 3})
        >>> len(index)
        3
        >>> directory = index.directory
        >>> inventory = Index(directory, d=4)
        >>> inventory['b']
        2
        >>> len(inventory)
        4
        """
        if args and isinstance(args[0], (bytes, str)):
            directory = args[0]
            args = args[1:]
        else:
            if args and args[0] is None:
                args = args[1:]
            directory = None
        self._cache = Cache(directory, eviction_policy='none')
        self.update(*args, **kwargs)

    @classmethod
    def fromcache(cls, cache, *args, **kwargs):
        """Initialize index using `cache` and update items.
        >>> cache = Cache()
        >>> index = Index.fromcache(cache, {'a': 1, 'b': 2, 'c': 3})
        >>> index.cache is cache
        True
        >>> len(index)
        3
        >>> 'b' in index
        True
        >>> index['c']
        3
        :param Cache cache: cache to use
        :param args: mapping or sequence of items
        :param kwargs: mapping of items
        :return: initialized Index
        """
        # pylint: disable=no-member,protected-access
        self = cls.__new__(cls)
        self._cache = cache
        self.update(*args, **kwargs)
        return self

    @property
    def cache(self):
        "Cache used by index."
        return self._cache

    @property
    def directory(self):
        "Directory path where items are stored."
        return self._cache.directory

    def __getitem__(self, key):
        """index.__getitem__(key) <==> index[key]
        Return corresponding value for `key` in index.
        >>> index = Index()
        >>> index.update({'a': 1, 'b': 2})
        >>> index['a']
        1
        >>> index['b']
        2
        >>> index['c']
        Traceback (most recent call last):
            ...
        KeyError: 'c'
        :param key: key for item
        :return: value for item in index with given key
        :raises KeyError: if key is not found
        """
        return self._cache[key]

    def __setitem__(self, key, value):
        """index.__setitem__(key, value) <==> index[key] = value
        Set `key` and `value` item in index.
        >>> index = Index()
        >>> index['a'] = 1
        >>> index[0] = None
        >>> len(index)
        2
        :param key: key for item
        :param value: value for item
        """
        self._cache[key] = value

    def __delitem__(self, key):
        """index.__delitem__(key) <==> del index[key]
        Delete corresponding item for `key` from index.
        >>> index = Index()
        >>> index.update({'a': 1, 'b': 2})
        >>> del index['a']
        >>> del index['b']
        >>> len(index)
        0
        >>> del index['c']
        Traceback (most recent call last):
            ...
        KeyError: 'c'
        :param key: key for item
        :raises KeyError: if key is not found
        """
        del self._cache[key]

    def setdefault(self, key, default=None):
        """Set and get value for `key` in index using `default`.
        If `key` is not in index then set corresponding value to `default`. If
        `key` is in index then ignore `default` and return existing value.
        >>> index = Index()
        >>> index.setdefault('a', 0)
        0
        >>> index.setdefault('a', 1)
        0
        :param key: key for item
        :param default: value if key is missing (default None)
        :return: value for item in index with given key
        """
        _cache = self._cache
        while True:
            try:
                return _cache[key]
            except KeyError:
                _cache.add(key, default, retry=True)

    def peekitem(self, last=True):
        """Peek at key and value item pair in index based on iteration order.
        >>> index = Index()
        >>> for num, letter in enumerate('xyz'):
        ...     index[letter] = num
        >>> index.peekitem()
        ('z', 2)
        >>> index.peekitem(last=False)
        ('x', 0)
        :param bool last: last item in iteration order (default True)
        :return: key and value item pair
        :raises KeyError: if cache is empty
        """
        return self._cache.peekitem(last, retry=True)

    def pop(self, key, default=ENOVAL):
        """Remove corresponding item for `key` from index and return value.
        If `key` is missing then return `default`. If `default` is `ENOVAL`
        then raise KeyError.
        >>> index = Index({'a': 1, 'b': 2})
        >>> index.pop('a')
        1
        >>> index.pop('b')
        2
        >>> index.pop('c', default=3)
        3
        >>> index.pop('d')
        Traceback (most recent call last):
            ...
        KeyError: 'd'
        :param key: key for item
        :param default: return value if key is missing (default ENOVAL)
        :return: value for item if key is found else default
        :raises KeyError: if key is not found and default is ENOVAL
        """
        _cache = self._cache
        value = _cache.pop(key, default=default, retry=True)
        if value is ENOVAL:
            raise KeyError(key)
        return value

    def popitem(self, last=True):
        """Remove and return item pair.
        Item pairs are returned in last-in-first-out (LIFO) order if last is
        True else first-in-first-out (FIFO) order. LIFO order imitates a stack
        and FIFO order imitates a queue.
        >>> index = Index()
        >>> index.update([('a', 1), ('b', 2), ('c', 3)])
        >>> index.popitem()
        ('c', 3)
        >>> index.popitem(last=False)
        ('a', 1)
        >>> index.popitem()
        ('b', 2)
        >>> index.popitem()
        Traceback (most recent call last):
          ...
        KeyError: 'dictionary is empty'
        :param bool last: pop last item pair (default True)
        :return: key and value item pair
        :raises KeyError: if index is empty
        """
        # pylint: disable=arguments-differ,unbalanced-tuple-unpacking
        _cache = self._cache

        with _cache.transact(retry=True):
            key, value = _cache.peekitem(last=last)
            del _cache[key]

        return key, value

    def push(self, value, prefix=None, side='back'):
        """Push `value` onto `side` of queue in index identified by `prefix`.
        When prefix is None, integer keys are used. Otherwise, string keys are
        used in the format "prefix-integer". Integer starts at 500 trillion.
        Defaults to pushing value on back of queue. Set side to 'front' to push
        value on front of queue. Side must be one of 'back' or 'front'.
        See also `Index.pull`.
        >>> index = Index()
        >>> print(index.push('apples'))
        500000000000000
        >>> print(index.push('beans'))
        500000000000001
        >>> print(index.push('cherries', side='front'))
        499999999999999
        >>> index[500000000000001]
        'beans'
        >>> index.push('dates', prefix='fruit')
        'fruit-500000000000000'
        :param value: value for item
        :param str prefix: key prefix (default None, key is integer)
        :param str side: either 'back' or 'front' (default 'back')
        :return: key for item in cache
        """
        return self._cache.push(value, prefix, side, retry=True)

    def pull(self, prefix=None, default=(None, None), side='front'):
        """Pull key and value item pair from `side` of queue in index.
        When prefix is None, integer keys are used. Otherwise, string keys are
        used in the format "prefix-integer". Integer starts at 500 trillion.
        If queue is empty, return default.
        Defaults to pulling key and value item pairs from front of queue. Set
        side to 'back' to pull from back of queue. Side must be one of 'front'
        or 'back'.
        See also `Index.push`.
        >>> index = Index()
        >>> for letter in 'abc':
        ...     print(index.push(letter))
        500000000000000
        500000000000001
        500000000000002
        >>> key, value = index.pull()
        >>> print(key)
        500000000000000
        >>> value
        'a'
        >>> _, value = index.pull(side='back')
        >>> value
        'c'
        >>> index.pull(prefix='fruit')
        (None, None)
        :param str prefix: key prefix (default None, key is integer)
        :param default: value to return if key is missing
            (default (None, None))
        :param str side: either 'front' or 'back' (default 'front')
        :return: key and value item pair or default if queue is empty
        """
        return self._cache.pull(prefix, default, side, retry=True)

    def clear(self):
        """Remove all items from index.
        >>> index = Index({'a': 0, 'b': 1, 'c': 2})
        >>> len(index)
        3
        >>> index.clear()
        >>> dict(index)
        {}
        """
        self._cache.clear(retry=True)

    def __iter__(self):
        """index.__iter__() <==> iter(index)
        Return iterator of index keys in insertion order.
        """
        return iter(self._cache)

    def __reversed__(self):
        """index.__reversed__() <==> reversed(index)
        Return iterator of index keys in reversed insertion order.
        >>> index = Index()
        >>> index.update([('a', 1), ('b', 2), ('c', 3)])
        >>> iterator = reversed(index)
        >>> next(iterator)
        'c'
        >>> list(iterator)
        ['b', 'a']
        """
        return reversed(self._cache)

    def __len__(self):
        """index.__len__() <==> len(index)
        Return length of index.
        """
        return len(self._cache)

    def keys(self):
        """Set-like object providing a view of index keys.
        >>> index = Index()
        >>> index.update({'a': 1, 'b': 2, 'c': 3})
        >>> keys_view = index.keys()
        >>> 'b' in keys_view
        True
        :return: keys view
        """
        return KeysView(self)

    def values(self):
        """Set-like object providing a view of index values.
        >>> index = Index()
        >>> index.update({'a': 1, 'b': 2, 'c': 3})
        >>> values_view = index.values()
        >>> 2 in values_view
        True
        :return: values view
        """
        return ValuesView(self)

    def items(self):
        """Set-like object providing a view of index items.
        >>> index = Index()
        >>> index.update({'a': 1, 'b': 2, 'c': 3})
        >>> items_view = index.items()
        >>> ('b', 2) in items_view
        True
        :return: items view
        """
        return ItemsView(self)

    __hash__ = None  # type: ignore

    def __getstate__(self):
        return self.directory

    def __setstate__(self, state):
        self.__init__(state)

    def __eq__(self, other):
        """index.__eq__(other) <==> index == other
        Compare equality for index and `other`.
        Comparison to another index or ordered dictionary is
        order-sensitive. Comparison to all other mappings is order-insensitive.
        >>> index = Index()
        >>> pairs = [('a', 1), ('b', 2), ('c', 3)]
        >>> index.update(pairs)
        >>> from collections import OrderedDict
        >>> od = OrderedDict(pairs)
        >>> index == od
        True
        >>> index == {'c': 3, 'b': 2, 'a': 1}
        True
        :param other: other mapping in equality comparison
        :return: True if index equals other
        """
        if len(self) != len(other):
            return False

        if isinstance(other, (Index, OrderedDict)):
            alpha = ((key, self[key]) for key in self)
            beta = ((key, other[key]) for key in other)
            pairs = zip(alpha, beta)
            return not any(a != x or b != y for (a, b), (x, y) in pairs)
        else:
            return all(self[key] == other.get(key, ENOVAL) for key in self)

    def __ne__(self, other):
        """index.__ne__(other) <==> index != other
        Compare inequality for index and `other`.
        Comparison to another index or ordered dictionary is
        order-sensitive. Comparison to all other mappings is order-insensitive.
        >>> index = Index()
        >>> index.update([('a', 1), ('b', 2), ('c', 3)])
        >>> from collections import OrderedDict
        >>> od = OrderedDict([('c', 3), ('b', 2), ('a', 1)])
        >>> index != od
        True
        >>> index != {'a': 1, 'b': 2}
        True
        :param other: other mapping in inequality comparison
        :return: True if index does not equal other
        """
        return not self == other

    def memoize(self, name=None, typed=False):
        """Memoizing cache decorator.
        Decorator to wrap callable with memoizing function using cache.
        Repeated calls with the same arguments will lookup result in cache and
        avoid function evaluation.
        If name is set to None (default), the callable name will be determined
        automatically.
        If typed is set to True, function arguments of different types will be
        cached separately. For example, f(3) and f(3.0) will be treated as
        distinct calls with distinct results.
        The original underlying function is accessible through the __wrapped__
        attribute. This is useful for introspection, for bypassing the cache,
        or for rewrapping the function with a different cache.
        >>> from diskcache import Index
        >>> mapping = Index()
        >>> @mapping.memoize()
        ... def fibonacci(number):
        ...     if number == 0:
        ...         return 0
        ...     elif number == 1:
        ...         return 1
        ...     else:
        ...         return fibonacci(number - 1) + fibonacci(number - 2)
        >>> print(fibonacci(100))
        354224848179261915075
        An additional `__cache_key__` attribute can be used to generate the
        cache key used for the given arguments.
        >>> key = fibonacci.__cache_key__(100)
        >>> print(mapping[key])
        354224848179261915075
        Remember to call memoize when decorating a callable. If you forget,
        then a TypeError will occur. Note the lack of parenthenses after
        memoize below:
        >>> @mapping.memoize
        ... def test():
        ...     pass
        Traceback (most recent call last):
            ...
        TypeError: name cannot be callable
        :param str name: name given for callable (default None, automatic)
        :param bool typed: cache different types separately (default False)
        :return: callable decorator
        """
        return self._cache.memoize(name, typed)

    @contextmanager
    def transact(self):
        """Context manager to perform a transaction by locking the index.
        While the index is locked, no other write operation is permitted.
        Transactions should therefore be as short as possible. Read and write
        operations performed in a transaction are atomic. Read operations may
        occur concurrent to a transaction.
        Transactions may be nested and may not be shared between threads.
        >>> from diskcache import Index
        >>> mapping = Index()
        >>> with mapping.transact():  # Atomically increment two keys.
        ...     mapping['total'] = mapping.get('total', 0) + 123.4
        ...     mapping['count'] = mapping.get('count', 0) + 1
        >>> with mapping.transact():  # Atomically calculate average.
        ...     average = mapping['total'] / mapping['count']
        >>> average
        123.4
        :return: context manager for use in `with` statement
        """
        with self._cache.transact(retry=True):
            yield

    def __repr__(self):
        """index.__repr__() <==> repr(index)
        Return string with printable representation of index.
        """
        name = type(self).__name__
        return '{0}({1!r})'.format(name, self.directory)


from typing import Dict, Any, Union, Type

from .core import run_fuze, BaseFuzerCls, MountPoint
from .cloud import GCSFuze, S3Fuze, MinioFuze, S3CompatFuze

FuzeTypes = Union[type(BaseFuzerCls), GCSFuze, S3Fuze, MinioFuze, S3CompatFuze]

_URI_PREFIXES_TO_CLS: Dict[str, FuzeTypes] = {
    'gs://': GCSFuze,
    's3://': S3Fuze,
    'minio://': MinioFuze,
    's3compat://': S3CompatFuze,
}

def get_fuze(source: str) -> FuzeTypes:
    uri_splits = source.split('://', maxsplit=1)
    if len(uri_splits) > 1:
        return _URI_PREFIXES_TO_CLS[uri_splits[0] + '://']
    return BaseFuzerCls


def autofuze_mount(source: str, mount_path: str, ready_file: bool = True, foreground: bool = False, threads: bool = False, fs_args: Dict[str, Any] = {}, cleanup: bool = True, *args, **kwargs) -> FuzeTypes:
    """ 
    Mounts a source path/bucket/etc with this FuzeCls at mount_path

    source: Bucket/path that will be the source of the mount
    mount_path: the directory that will be mounted to
    foreground: return a thread blocking function, otherwise will be a background thread that will need to be killed explicitly.
    fs_args: Dict[str, Any] that will be passed when initializing the FileSystem (such as auth)
    cleanup: removes the mount_path on unmount
    """
    uri_splits = source.split('://', maxsplit=1)
    if len(uri_splits) > 1:
        fuzecls = _URI_PREFIXES_TO_CLS[uri_splits[0] + '://']
        source_path = uri_splits[-1]
    else:
        fuzecls = BaseFuzerCls
        source_path = source
    fuzecls.mount(source_path, mount_path, ready_file = ready_file, foreground = foreground, threads = threads, fs_args = fs_args, cleanup = cleanup, *args, **kwargs)
    return fuzecls


def autofuze_unmount(source: str, mount_path: str, timeout: int = 5, force: bool = False, *args, **kwargs) -> FuzeTypes:
    uri_splits = source.split('://', maxsplit=1)
    if len(uri_splits) > 1:
        fuzecls = _URI_PREFIXES_TO_CLS[uri_splits[0] + '://']
    else:
        fuzecls = BaseFuzerCls
    fuzecls.unmount(mount_path, timeout = timeout, force = force)
    return fuzecls


from . import static
from .core import (
    DEFAULT_SETTINGS,
    ENOVAL,
    EVICTION_POLICY,
    UNKNOWN,
    Cache,
    Disk,
    EmptyDirWarning,
    JSONDisk,
    Timeout,
    UnknownFileWarning,
)

from ._json import OrJSONDisk


__all__ = (
    'TimeValues',
    'DefaultHeaders'
)

from typing import Dict, Any

TimeValues: Dict[str, int] = {
    'secs': 60,
    'mins': 60,
    'hrs': 60,
    'days': 24,
    'wks': 7,
    'mnths': 4
}


DefaultHeaders: Dict[str, Any] = {
    'Accept': 'application/json',
    'Content-Type': 'application/json'
}




__all__ = (
    'BaseModel',
    'Field',
    'BaseCls',
    'BaseDataCls',
    'BaseLazy',
    'LazyCls',
    'create_lazycls',
    'clear_lazy_models',
    'set_modulename',
)

from . import utils
from . import base
from . import core
from . import timez

from .base import *
from .core import *



import orjson as _orjson
import ujson as _ujson
import simdjson as _simdjson
import rapidjson as _rjson

from .static import *
from .core import Disk, JSONDisk, _zlib

_parser = _simdjson.Parser()

## OrJSON is only stable for non-float/int types like strings
## Since we already use orjson, we include this

class OrJSONDisk(Disk):
    "Cache key and value using JSON serialization with _zlib compression."

    def __init__(self, directory, compress_level=5, **kwargs):
        """Initialize JSON disk instance.
        Keys and values are compressed using the _zlib library. The
        `compress_level` is an integer from 0 to 9 controlling the level of
        compression; 1 is fastest and produces the least compression, 9 is
        slowest and produces the most compression, and 0 is no compression.
        :param str directory: directory path
        :param int compress_level: _zlib compression level (default 5)
        :param kwargs: super class arguments
        """
        self.compress_level = compress_level
        super().__init__(directory, **kwargs)

    def put(self, key):
        json_bytes = _orjson.dumps(key)
        data = _zlib.compress(json_bytes, self.compress_level)
        return super().put(data)

    def get(self, key, raw):
        data = super().get(key, raw)
        return _orjson.loads(_zlib.decompress(data))

    def store(self, value, read, key=UNKNOWN):
        if not read:
            json_bytes = _orjson.dumps(value)
            value = _zlib.compress(json_bytes, self.compress_level)
        return super().store(value, read, key=key)

    def fetch(self, mode, filename, value, read):
        data = super().fetch(mode, filename, value, read)
        if not read:
            data = _orjson.loads(_zlib.decompress(data))
        return data




from pydantic import create_model
from .types import *
from .prop import classproperty
from .models import BaseLazy, BaseModel, BaseCls

_GeneratedLazyModels = {}
_BaseModuleName = 'lazycls'

def clear_lazy_models():
    """
    Function to Clear out any Generated Lazy Models
    """
    global _GeneratedLazyModels
    _GeneratedLazyModels = {}

def set_modulename(name: str = 'lazycls'):
    """
    Function to Set the Module Name. Useful when using this within other projects.
    """
    global _BaseModuleName
    _BaseModuleName = name


def generate_modelcls(_clsname: str, data: Dict[str, Any], modulename: str = _BaseModuleName, basecls: Type[BaseModel] = BaseLazy):
    """
    Primary Function to Generate the Model Class Dynamically
    """
    global _GeneratedLazyModels
    if _clsname not in _GeneratedLazyModels:
        clsdata = {k: type(v) for k,v in data.items()}
        _GeneratedLazyModels[_clsname] = create_model(_clsname, __base__ = basecls, __module__ = modulename, **clsdata)
    return _GeneratedLazyModels[_clsname](**data)

def create_lazycls(clsname: str, data: Dict[str, Any], modulename: str = _BaseModuleName, basecls: Type[BaseModel] = BaseLazy) -> Type[BaseCls]:
    """
    Iterator Function to take data and recurisvely turn them into LazyCls Models
    """
    if clsname.endswith('s'): clsname = clsname[:-1]
    if isinstance(data, str): return data
    for k,v in data.items():
        subcls = f'{clsname}_{k}'
        if v and isinstance(v, list):
            data[k] = [generate_modelcls(_clsname=subcls, data=i, modulename=modulename, basecls=basecls) if isinstance(i, dict) else i  for i in v]
        elif isinstance(v, dict):
            for a,b in v.items():
                if isinstance(b, dict):
                    subsubcls = f'{clsname}_{k}_{a}'
                    v[a] = generate_modelcls(_clsname=subsubcls, data=b, modulename=modulename, basecls=basecls)
            data[k] = generate_modelcls(_clsname=subcls, data=v, modulename=modulename, basecls=basecls)
    return generate_modelcls(_clsname=clsname, data=data, modulename=modulename, basecls=basecls)


class LazyCls:

    @classmethod
    def create(cls, *args, **kwargs):
        return create_lazycls(*args, **kwargs)
    
    @classmethod
    def get(cls, name):
        return _GeneratedLazyModels.get(name)
    
    @classproperty
    def models(cls):
        return _GeneratedLazyModels

    @classproperty
    def modelNames(cls):
        return list(cls.models.keys())

    @classproperty
    def __len__(cls):
        return len(_GeneratedLazyModels)

    @classmethod
    def __getitem__(cls, name):
        return _GeneratedLazyModels.get(name)
    
    @classmethod
    def __call__(cls, name, data: Dict[str, Any], *args, **kwargs):
        return create_lazycls(clsname=name, data=data, *args, **kwargs)


__all__ = [
    'LazyCls',
    'create_lazycls',
    'clear_lazy_models',
    'set_modulename'
]


from .base import *

def is_iterable(value):
    try: iter(value)
    except TypeError: return False
    return True


class ExecFile:
    """ Simple container for a filename. Mainly needed to be able to run
        `isinstance(..., ExecFile)`
        Extends with some utils from modified lazycls.ext.pathio
    """

    def __init__(self, filename):
        self.filename = filename
        self._path = None
    
    @property
    def path(self) -> PathLike:
        if not self._path: self._path = Path.get_path(self.filename, resolve=True)
        return self._path
    
    @property
    def is_file(self): return self.path.is_file()
    
    @property
    def is_dir(self): return self.path.is_dir()
    
    @property
    def exists(self): return self.path.exists()
    
    
    


import io
import inspect
import reprlib
import pathlib

import types

from copy import copy
from glob import glob
from lazycls.prop import classproperty
from lazycls.serializers import OrJson
from subprocess import PIPE, Popen, TimeoutExpired
from lazycls.ext.pathio import Path, PathLike




from . import base
from . import static
from . import prop
from . import pyd


from .base import *
from .static import *
from .prop import classproperty, ClasspropertyMeta


__all__ = (
    'PathLike',
    'TYPE_CHECKING',
    'List', 'Dict', 'AnyStr', 'Any',
    'Optional', 'Union', 'Tuple', 'Mapping', 'Sequence', 'TypeVar', 'Type',
    'Callable', 'Coroutine', 'Generator', 'IO', 'Iterable', 'Iterator', 'AsyncIterator',
    'Data', 'AnyMany', 'TextMany', 'TextList',
    'DictList', 'DictMany', 'DictAny', 'DictAny',
    'DefaultHeaders', 'TimeValues',
    'classproperty', 'ClasspropertyMeta'
)




"""
Custom Types for lazycls.ext.envcls
"""
import os
from .prop import classproperty

""" For custom types """
# https://pydantic-docs.helpmanual.io/usage/types/#classes-with-__get_validators__

class EnvType(str):
    """
    Base Env Type that uses str
    """
    @classmethod
    def __get_validators__(cls): yield cls.validate
    
    @classmethod
    def cast(cls, v: str): return v
    
    @classmethod
    def cast_to_env(cls, v): return v
    
    
    @classproperty
    def default_value(cls): return ""
    @classproperty
    def list_delimiter(cls): return ','
    @classproperty
    def dict_delimiter(cls): return '='
    @classproperty
    def true_values(cls): return {'true', 'True', '1', 'yes', 'Yes'}
    @classproperty
    def false_values(cls): return {'false', 'False', '0', 'no', 'No'}
    @classproperty
    def none_values(cls): return {'', 'none', 'None', 'Null', 'null', 'n/a', 'N/A'}
    
    @classproperty
    def to_env_key(cls): return None
    @classproperty
    def to_env_default_value(cls): return None
    @classproperty
    def override_env_to_value(cls) -> bool: return False
    
    @classmethod
    def get_default_value(cls): return cls.default_value
    
    @classmethod
    def get_to_env_value(cls): 
        if not cls.to_env_key: return None
        return os.getenv(cls.to_env_key, cls.to_env_default_value)
    
    @classmethod
    def set_to_env(cls, to_val):
        if not to_val: return
        cur_env_val = cls.get_to_env_value()
        if cur_env_val and cur_env_val is not cls.to_env_default_value or cls.override_env_to_value:
            os.environ[cls.to_env_key] = to_val
    
    @classmethod
    def validate(cls, v):
        if not v: return cls.get_default_value()
        try: 
            val = cls.cast(v)
            if cls.to_env_key: 
                to_val = cls.cast_to_env(val)
                cls.set_to_env(to_val)
                return to_val
            return val
        except Exception as e: 
            #print(e)
            return ""
    
    @classmethod
    def __modify_schema__(cls, field_schema): pass
    
    def __repr__(self): return f'{self.__class__.__name__}({super().__repr__()})'
    

class ListStr(EnvType):
    """
    Returns List[str] by splitting on delimiter ','
    """
    @classproperty
    def default_value(cls): return []
    
    @classmethod
    def cast(cls, v: str):
        return v.split(cls.list_delimiter)
    
class DictStr(EnvType):
    """
    Returns List[str] by splitting on delimiter ','
    """
    @classproperty
    def default_value(cls): return {}
    
    @classmethod
    def cast(cls, v: str):
        val = v.split(cls.list_delimiter)
        if not val: return cls.default_value
        rez = {}
        for i in val:
            for k,v in i.split(cls.dict_delimiter, 1):
                rez[k.strip()] = v.strip()
        return rez

from lazycls.ext._srlz import Srlz

class Base64(EnvType):
    """
    Returns Base64 Encoded Strings
    """
    @classmethod
    def cast(cls, v: str): return Srlz.Base.b64_decode(v)

class Base64Gzip(EnvType):
    """
    Returns Base64 + Gzip Encoded Strings
    """
    @classmethod
    def cast(cls, v: str): return Srlz.Base.b64_gzip_decode(v)


class YamlStr(EnvType):
    """
    Returns Yaml.loads(str)
    """
    @classproperty
    def default_value(cls): return {}
    
    @classmethod
    def cast(cls, v: str): return Srlz.Yaml.loads(v)


class PathStr(EnvType):
    """
    Returns Path.get_path(str, resolve=True)
    """
    @classproperty
    def default_value(cls): return None
    
    @classmethod
    def cast(cls, v: str): return Path.get_path(v, resolve=True)


class JsonB64Str(EnvType):
    """
    Returns Json.loads(Base.b64_decode(str))
    """
    @classproperty
    def default_value(cls): return {}
    
    @classmethod
    def cast(cls, v: str): 
        return Srlz.Json.loads(Srlz.Base.b64_decode(v))

class JsonB64GZipStr(EnvType):
    """
    Returns Json.loads(Base.b64_gzip_decode(str))
    """
    @classproperty
    def default_value(cls): return {}
    
    @classmethod
    def cast(cls, v: str): 
        return OrJson.loads(Base.b64_gzip_decode(v))


class AuthzFileStr(EnvType):
    """
    Returns Path to Credentials File
    """
    @classproperty
    def to_env_key(cls): return 'AUTHZ_CREDENTIALS'
    @classproperty
    def to_env_default_value(cls): return None
    @classproperty
    def to_env_path(cls):
        if LazyLib.is_avail_colab: return Path.get_path('/authz/auth_file.txt')
        return Path.cwd().joinpath('authz', 'auth_file.txt')
    
    @classmethod
    def dump_decoded(cls, v): return v
    
    @classmethod
    def write_to_file(cls, data):
        cls.to_env_path.parent.mkdir(exist_ok=True, parents=True)
        cls.to_env_path.write_text(data, encoding='utf-8') 
        return cls.to_env_path.string

    @classmethod
    def cast_to_env(cls, v):
        cur_env_val = cls.get_to_env_value()
        if cur_env_val and Path(cur_env_val).exists() and not cls.override_env_to_value: return cur_env_val
        data = cls.dump_decoded(v)
        if data: return cls.write_to_file(data)
        return cur_env_val

class GoogleAuthJsonStr(AuthzFileStr):
    """
    Returns Path to GOOGLE_APPLICATION_CREDENTIALS
    """
    @classproperty
    def to_env_key(cls): return 'GOOGLE_APPLICATION_CREDENTIALS'
    @classproperty
    def to_env_path(cls):
        if LazyLib.is_avail_colab: return Path.get_path('/authz/adc.json')
        return Path.cwd().joinpath('authz', 'adc.json')
    @classmethod
    def cast(cls, v: str): return OrJson.loads(v)
    @classmethod
    def dump_decoded(cls, v): return OrJson.dumps(v)
    
class GoogleAuthB64(GoogleAuthJsonStr):
    """
    Returns Path to GOOGLE_APPLICATION_CREDENTIALS
    """
    
    @classmethod
    def cast(cls, v: str): return OrJson.loads(Base.b64_decode(v))
    
class GoogleAuthBGZ(GoogleAuthJsonStr):
    """
    Returns Path to GOOGLE_APPLICATION_CREDENTIALS
    """
    @classmethod
    def cast(cls, v: str): return OrJson.loads(Base.b64_gzip_decode(v))




import os
import stat as stat_  # aliasing because there's a 'stat' UNIX command
import string
from .base import *
from .core import Cmd


def overload_chars(locals_):
    """ Assigns all ascii characters as values to keys of the same name in the
        `locals_` argument. Intended to overload the `locals()` call of modules
        that import `lazycls.ext._cmds.auto`
            >>> lazycls.ext._cmds.auto import overload_chars, ls
            >>> overload_chars(locals())
            >>> ls -l
            <<< -rw-r--r-- 1 kbairak kbairak 9401 Feb  3 23:12 foo.txt
            ... -rw-r--r-- 1 kbairak kbairak 8923 Feb  3 23:06 bar.txt
    """

    for char in string.ascii_letters:
        if char in locals_: continue
        locals_[char] = char

for path in os.get_exec_path():
    try: listdir = os.listdir(path)
    except FileNotFoundError: continue
    for original_name in listdir:
        name = original_name.replace('-', '_')
        if name in locals(): continue        
        if 'x' in stat_.filemode(os.lstat(os.path.join(path, original_name)).st_mode):
            locals()[name] = Cmd(original_name)



__all__ = (
    'full_name',
    'WindowsError',
    'Constant',
    'DBNAME',
    'ENOVAL',
    'UNKNOWN',
    'MODE_NONE',
    'MODE_RAW',
    'MODE_BINARY',
    'MODE_TEXT',
    'MODE_PICKLE',
    'DEFAULT_SETTINGS',
    'METADATA',
    'EVICTION_POLICY'
)

import dill

def full_name(func):
    "Return full name of `func` by adding the module and function name."
    return func.__module__ + '.' + func.__qualname__



try: WindowsError
except NameError:
    class WindowsError(Exception):
        "Windows error place-holder on platforms without support."


class Constant(tuple):
    "Pretty display of immutable constant."

    def __new__(cls, name):
        return tuple.__new__(cls, (name,))

    def __repr__(self):
        return '%s' % self[0]


DBNAME = 'cache.db'
ENOVAL = Constant('ENOVAL')
UNKNOWN = Constant('UNKNOWN')

MODE_NONE = 0
MODE_RAW = 1
MODE_BINARY = 2
MODE_TEXT = 3
MODE_PICKLE = 4

DEFAULT_SETTINGS = {
    u'statistics': 0,  # False
    u'tag_index': 0,  # False
    u'eviction_policy': u'least-recently-stored',
    u'size_limit': 2 ** 30,  # 1gb
    u'cull_limit': 10,
    u'sqlite_auto_vacuum': 1,  # FULL
    u'sqlite_cache_size': 2 ** 13,  # 8,192 pages
    u'sqlite_journal_mode': u'wal',
    u'sqlite_mmap_size': 2 ** 26,  # 64mb
    u'sqlite_synchronous': 1,  # NORMAL
    u'disk_min_file_size': 2 ** 15,  # 32kb
    u'disk_pickle_protocol': dill.HIGHEST_PROTOCOL,
}

METADATA = {
    u'count': 0,
    u'size': 0,
    u'hits': 0,
    u'misses': 0,
}

EVICTION_POLICY = {
    'none': {
        'init': None,
        'get': None,
        'cull': None,
    },
    'least-recently-stored': {
        'init': (
            'CREATE INDEX IF NOT EXISTS Cache_store_time ON'
            ' Cache (store_time)'
        ),
        'get': None,
        'cull': 'SELECT {fields} FROM Cache ORDER BY store_time LIMIT ?',
    },
    'least-recently-used': {
        'init': (
            'CREATE INDEX IF NOT EXISTS Cache_access_time ON'
            ' Cache (access_time)'
        ),
        'get': 'access_time = {now}',
        'cull': 'SELECT {fields} FROM Cache ORDER BY access_time LIMIT ?',
    },
    'least-frequently-used': {
        'init': (
            'CREATE INDEX IF NOT EXISTS Cache_access_count ON'
            ' Cache (access_count)'
        ),
        'get': 'access_count = access_count + 1',
        'cull': 'SELECT {fields} FROM Cache ORDER BY access_count LIMIT ?',
    },
}




import os
import sys
from pathlib import Path
from setuptools import setup, find_packages

if sys.version_info.major != 3:
    raise RuntimeError("This package requires Python 3+")

version = '0.1.03'
pkg_name = 'lazycls'
gitrepo = 'trisongz/lazycls'
root = Path(__file__).parent

requirements = [
    #'aiofiles',
    'anyio',
    'dill',
    'fsspec',
    'orjson',
    #'pickletools',
    'pydantic[dotenv]',
    'pylogz',
    'pysimdjson',
    'pyyaml',
    'sniffio',
    'universal_pathlib'
]

extras = {
    'gcs': ['gcsfs'],
    's3': ['s3fs'], 
    'cloudfs': ['gcsfs', 's3fs'],
}

args = {
    'packages': find_packages(include = ['lazycls', 'lazycls.*', 'lazy', 'lazy.*']),
    'install_requires': requirements,
    'include_package_data': True,
    'long_description': root.joinpath('README.md').read_text(encoding='utf-8'),
    'entry_points': {}
}

setup(
    name=pkg_name,
    version=version,
    url=f'https://github.com/{gitrepo}',
    license='MIT Style',
    description='Dynamic Dataclasses for the Super Lazy',
    author='Tri Songz',
    author_email='ts@growthengineai.com',
    long_description_content_type="text/markdown",
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries',
    ],
    **args
)


import os
import sys
import subprocess
import logging
import platform
from pathlib import Path
from typing import Union, List, Type


logger = logging.getLogger(name='lazycls')


def exec_cmd(cmd, raise_error: bool = True):
    try:
        out = subprocess.check_output(cmd, shell=True)
        if isinstance(out, bytes): out = out.decode('utf8')
        return out.strip()
    except Exception as e:
        if not raise_error: return ""
        raise e

def exec_daemon(cmd: Union[List[str], str], stdout = subprocess.PIPE, stderr = subprocess.STDOUT, set_proc_uid: bool = True, *args, **kwargs):
    if isinstance(cmd, str): cmd = [cmd]
    if set_proc_uid and platform.system() != 'Darwin': return subprocess.Popen(cmd, stdout = stdout, stderr = stderr, preexec_fn = lambda: os.setuid(1), *args, **kwargs)
    return subprocess.Popen(cmd, stdout = stdout, stderr = stderr, *args, **kwargs)

def exec_run(cmd: Union[List[str], str], *args, **kwargs):
    if isinstance(cmd, str): cmd = [cmd]
    return subprocess.run(cmd, *args, **kwargs)

def exec_out(cmd: Union[List[str], str], shell: bool = True, *args, **kwargs):
    if isinstance(cmd, str): cmd = [cmd]
    return subprocess.check_output(cmd, shell = shell, *args, **kwargs)


def exec_shell(cmd): return os.system(cmd)
def getParentPath(p: str) -> Path: return Path(p).parent
def to_camelcase(string: str) -> str: return ''.join(word.capitalize() for word in string.split('_'))

def toPath(path: Union[str, Path], resolve: bool = True) -> Type[Path]:
    if isinstance(path, str): path = Path(path)
    if resolve: path.resolve()
    return path

def get_user_path(path: Union[str, Path], resolve: bool = False) -> Type[Path]:
    if isinstance(path, str): path = Path(path)
    path = path.expanduser()
    if resolve: path.resolve()
    return path


def get_cwd(*paths, posix: bool = True) -> Union[str, Path]:
    if not paths:
        if posix: return Path.cwd().as_posix()
        return Path.cwd()
    if posix: return Path.cwd().joinpath(*paths).as_posix()
    return Path.cwd().joinpath(*paths)

def list_to_chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]

def list_to_dict(items, delim='='):
    res = {}
    for item in items:
        i = item.split(delim, 1)
        res[i[0].strip()] = i[-1].strip()
    return res


def get_variable_separator():
    """
    Returns the environment variable separator for the current platform.
    :return: Environment variable separator
    """
    if sys.platform.startswith('win'):
        return ';'
    return ':'


def find_binary_in_path(filename):
    """
    Searches for a binary named `filename` in the current PATH. If an executable is found, its absolute path is returned
    else None.
    :param filename: Filename of the binary
    :return: Absolute path or None
    """
    if 'PATH' not in os.environ:
        return None
    for directory in os.environ['PATH'].split(get_variable_separator()):
        binary = os.path.abspath(os.path.join(directory, filename))
        if os.path.isfile(binary) and os.access(binary, os.X_OK):
            return binary
    return None

def exec_sed(path: Union[str, Path], changes: List[set], new_path: Union[str, Path] = None, chmod: int = None) -> Path:
    """ Does a cleaner `sed` command 
        - path: path to file
        - changes: List of {find, replace, n_times}
        - new_path: new filepath, otherwise will write to original
        - chmod: int 
    """
    f = toPath(path)
    txt = f.read_text(encoding='utf-8')
    for c in changes:
        txt = txt.replace(str(c[0]), str(c[1]), c[2]) if len(c) == 3 else txt.replace(str(c[0]), str(c[1]))
    if new_path: f = toPath(new_path)
    f.write_text(txt, encoding='utf-8')
    if chmod: f.chmod(chmod)
    return f

# Aliases that wont be exported by default
run = exec_run
cmd = exec_cmd
exec_command = exec_cmd

get_parent_path = getParentPath
to_path = toPath


__all__ = [
    'exec_run',
    'exec_out',
    'exec_cmd',
    'exec_daemon',
    'exec_command',
    'exec_sed',
    'get_parent_path',
    'getParentPath',
    'to_path',
    'to_user_path',
    'get_cwd',
    'toPath',
    'to_camelcase',
]

